﻿using System.Xml;
using System.Collections.Generic;
using System;
using System.Collections;
using System.IO;

namespace Persistence {
  public class XmlParser : IEnumerable<IParseToken> {

    private XmlReader reader;
    private delegate IEnumerator<IParseToken> Handler();
    private Dictionary<string, Handler> handlers;

    public XmlParser(StreamReader input) {
      XmlReaderSettings settings = new XmlReaderSettings();
      settings.ConformanceLevel = ConformanceLevel.Document;
      settings.IgnoreWhitespace = true;
      settings.IgnoreComments = true;
      this.reader = XmlReader.Create(input, settings);
      this.handlers = new Dictionary<string, Handler>();
      this.handlers.Add("PRIMITIVE", new Handler(ParsePrimitive));
      this.handlers.Add("COMPOSITE", new Handler(ParseComposite));
      this.handlers.Add("REFERENCE", new Handler(ParseReference));
      this.handlers.Add("NULL", new Handler(ParseNull));
    }
    public IEnumerator<IParseToken> GetEnumerator() {
      while (this.reader.Read()) {
        if (!reader.IsStartElement()) {
          break;
        }
        IEnumerator<IParseToken> iterator;
        try {
          iterator = handlers[reader.Name].Invoke();
        }
        catch (KeyNotFoundException) {
          throw new InvalidOperationException(String.Format(
            "No handler for XML tag \"{0}\" installed",
            reader.Name));
        }
        while (iterator.MoveNext()) {
          yield return iterator.Current;
        }
      }
    }
    private IEnumerator<IParseToken> ParsePrimitive() {
      yield return new Primitive(
        this.reader.GetAttribute("name"),
        Type.GetType(this.reader.GetAttribute("type")),
        this.reader.ReadString());
    }
    private IEnumerator<IParseToken> ParseComposite() {
      string name = this.reader.GetAttribute("name");
      Type type = Type.GetType(this.reader.GetAttribute("type"));
      int id = int.Parse(this.reader.GetAttribute("id"));
      yield return new CompositeStart(name, type, id);
      IEnumerator<IParseToken> iterator = this.GetEnumerator();
      while (iterator.MoveNext())
        yield return iterator.Current;
      yield return new CompositeEnd(name, type, id);
    }
    private IEnumerator<IParseToken> ParseReference() {
      yield return new Reference(
        this.reader.GetAttribute("name"),
        int.Parse(this.reader.GetAttribute("ref")));
    }
    private IEnumerator<IParseToken> ParseNull() {
      yield return new Null(this.reader.GetAttribute("name"));
    }
    IEnumerator IEnumerable.GetEnumerator() {
      return this.GetEnumerator();
    }
  }  
}
